from collections import defaultdict
from typing import (
    overload,
    AsyncGenerator,
    Generic,
    Optional,
    Union,
    TypeVar,
    DefaultDict,
    Iterator,
    MutableMapping,
)

from .queuedist import QueueDist

_KT = TypeVar("_KT")
_VT = TypeVar("_VT")
_T = TypeVar("_T")


class SubscribableDict(MutableMapping[_KT, _VT], Generic[_KT, _VT]):
    """Dict with async generator that yields values as they change"""

    def __init__(self):
        self._map: DefaultDict[_KT, QueueDist[_VT]] = defaultdict(QueueDist)

    def subscribe(
        self,
        key: _KT,
        last: Optional[bool] = True,
    ) -> AsyncGenerator[_VT, None]:
        """Async generator that yields values for the key as they are set

        Waits for new values and yields them as they are set. If `last` is
        true, yields immediately the current value for the key before starting
        to wait. If the key doesn't exist, waits for the first value for the
        key; KeyError won't be raised.

        """
        return self._map[key].subscribe(last=last)

    def __getitem__(self, key: _KT) -> _VT:
        """The current value for the key. KeyError if the key doesn't exist."""
        if q := self._map.get(key):
            return q.get()
        else:
            raise KeyError

    @overload
    def get(self, key: _KT) -> Optional[_VT]:
        ...

    @overload
    def get(self, key: _KT, default: Union[_VT, _T]) -> Union[_VT, _T]:
        ...

    def get(self, key, default=None):
        """The current value for the key if the key exist, else the default"""
        if q := self._map.get(key):
            return q.get()
        else:
            return default

    def __setitem__(self, key: _KT, value: _VT) -> None:
        """Set the value for the key, yielding the value in the generators"""
        self._map[key].put(value)

    def __delitem__(self, key: _KT) -> None:
        """Remove the key, ending all subscriptions for the key

        The async generators returned by the method `subscribe()` for the key
        will return.

        KeyError will be raised if the key doesn't exist
        """
        self._map.pop(key).close()

    def close(self) -> None:
        """Remove all keys, ending all subscriptions for all keys"""
        while True:
            try:
                self.popitem()  # __delitem__() will be called
            except KeyError:
                break

    def __iter__(self) -> Iterator[_KT]:
        return iter(self._map)

    def __len__(self) -> int:
        return len(self._map)



import pytest

from nextline.state import Running, Exited, StateObsoleteError

from .base import BaseTestState


class TestRunning(BaseTestState):

    state_class = Running

    @pytest.fixture()
    def state(self, running):
        yield running

    async def assert_obsolete(self, state):
        assert "obsolete" in repr(state)

        with pytest.raises(StateObsoleteError):
            state.run()

        with pytest.raises(StateObsoleteError):
            await state.finish()

        with pytest.raises(StateObsoleteError):
            state.reset()

        with pytest.raises(StateObsoleteError):
            await state.close()

    @pytest.mark.asyncio
    async def test_exited(self, state):
        # exited() can be called multiple times
        exited = await state.exited()
        assert isinstance(exited, Exited)
        assert exited is await state.exited()
        assert exited is await state.exited()

        await self.assert_obsolete(state)

    def test_send_pdb_command(self, state):
        pass



from __future__ import annotations

import dataclasses
import datetime
import queue
from itertools import count

from ..utils import ThreadTaskDoneCallback, current_task_or_thread
from ..types import TraceInfo
from .ci import PdbCommandInterface
from .custom import CustomizedPdb
from .stream import StreamIn, StreamOut


from typing import (
    TYPE_CHECKING,
    Callable,
    Optional,
    Union,
    Any,
    Set,
    Dict,
    Tuple,
)
from types import FrameType

if TYPE_CHECKING:
    from ..types import TraceFunc
    from ..utils import SubscribableDict, ThreadTaskIdComposer


@dataclasses.dataclass(frozen=True)
class PdbCIState:
    prompting: int
    file_name: str
    line_no: int
    trace_event: str


def PdbInterfaceFactory(
    registry: SubscribableDict,
    pdb_ci_map: Dict[int, PdbCommandInterface],
    modules_to_trace: Set[str],
) -> Callable[[], PdbInterface]:

    id_composer: ThreadTaskIdComposer = registry["trace_id_factory"]
    trace_no_counter = count(1).__next__
    prompting_counter = count(1).__next__
    callback_map: Dict[Any, TraceInfo] = {}

    def callback_func(key):
        trace_info = callback_map[key]
        trace_no = trace_info.trace_no
        del registry[trace_no]
        nos = list(registry.get("trace_nos"))
        nos.remove(trace_no)
        nos = tuple(nos)
        registry["trace_nos"] = nos

        trace_info = dataclasses.replace(
            trace_info,
            state="finished",
            ended_at=datetime.datetime.now(),
        )
        registry["trace_info"] = trace_info

    callback = ThreadTaskDoneCallback(done=callback_func)

    registry["callback"] = callback
    # NOTE: Sending `callback` so `callback.close()` can be called when the
    # thread ends.

    def factory() -> PdbInterface:
        trace_no = trace_no_counter()

        registry[trace_no] = None
        nos = (registry.get("trace_nos") or ()) + (trace_no,)
        registry["trace_nos"] = nos

        run_no: int = registry["run_no"]
        thread_task_id = id_composer()

        task_or_thread = current_task_or_thread()
        registry["run_no_map"][task_or_thread] = run_no  # type: ignore
        registry["trace_no_map"][task_or_thread] = trace_no  # type: ignore

        trace_info = TraceInfo(
            run_no=run_no,
            trace_no=trace_no,
            thread_no=thread_task_id.thread_no,
            task_no=thread_task_id.task_no,
            state="running",
            started_at=datetime.datetime.now(),
        )
        registry["trace_info"] = trace_info

        key = callback.register()
        callback_map[key] = trace_info

        pbi = PdbInterface(
            trace_id=trace_no,
            registry=registry,
            ci_map=pdb_ci_map,
            prompting_counter=prompting_counter,
            modules_to_trace=modules_to_trace,
        )
        return pbi

    return factory


class PdbInterface:
    """Instantiate Pdb and register its command loops

    TODO: Update parameters

    Parameters
    ----------
    trace_id : object
        The Id to distiugish each instance of Pdb
    modules_to_trace: set
        The set of modules to trace. This object is shared by multiple
        objects. Modules in which Pdb commands are prompted will be
        added.
    registry: object
        A registry
    ci_map: object
        A registry
    prompting_counter : callable
        Used to count the Pdb command loops
    """

    def __init__(
        self,
        trace_id: int,
        registry: SubscribableDict,
        ci_map: Dict[int, PdbCommandInterface],
        prompting_counter: Callable[[], int],
        modules_to_trace: Set[str],
    ):
        self._trace_id = trace_id
        self._registry = registry
        self._ci_map = ci_map
        self._prompting_counter = prompting_counter
        self.modules_to_trace = modules_to_trace
        self._opened = False

        self._q_stdin: queue.Queue = queue.Queue()
        self._q_stdout: queue.Queue = queue.Queue()

        self._pdb = CustomizedPdb(
            pdbi=self,
            stdin=StreamIn(self._q_stdin),
            stdout=StreamOut(self._q_stdout),
            readrc=False,
        )

        self._trace_args: Optional[Tuple[FrameType, str, Any]] = None

    def trace(self, frame, event, arg) -> Optional[TraceFunc]:
        """Call Pdb while storing trace args"""

        def calling_trace(frame, event, arg) -> None:
            self._trace_args = (frame, event, arg)

        def exited_trace() -> None:
            self._trace_args = None

        def create_local_trace() -> TraceFunc:
            pdb_trace: Union[TraceFunc, None] = self._pdb.trace_dispatch

            def local_trace(frame, event, arg) -> Optional[TraceFunc]:
                nonlocal pdb_trace
                assert pdb_trace
                calling_trace(frame, event, arg)
                try:
                    if pdb_trace := pdb_trace(frame, event, arg):
                        return local_trace
                    return None
                finally:
                    exited_trace()

            return local_trace

        return create_local_trace()(frame, event, arg)

    def entering_cmdloop(self) -> None:
        """To be called by the custom Pdb before _cmdloop()"""

        if not self._trace_args:
            raise RuntimeError("calling_trace() must be called first")

        frame, event, _ = self._trace_args

        if module_name := frame.f_globals.get("__name__"):
            # TODO: This should be done somewhere else
            self.modules_to_trace.add(module_name)

        self._state = PdbCIState(
            prompting=self._prompting_counter(),
            file_name=self._pdb.canonic(frame.f_code.co_filename),
            line_no=frame.f_lineno,
            trace_event=event,
        )

        self._pdb_ci = PdbCommandInterface(
            pdb=self._pdb,
            queue_in=self._q_stdin,
            queue_out=self._q_stdout,
            counter=self._prompting_counter,
            trace_id=self._trace_id,
            registry=self._registry,
            trace_args=self._trace_args,
        )
        self._pdb_ci.start()

        self._ci_map[self._trace_id] = self._pdb_ci

        self._registry[self._trace_id] = self._state

    def exited_cmdloop(self) -> None:
        """To be called by the custom Pdb after _cmdloop()"""

        del self._ci_map[self._trace_id]

        self._state = dataclasses.replace(self._state, prompting=0)
        self._registry[self._trace_id] = self._state

        self._pdb_ci.end()



import asyncio
from pathlib import Path
from typing import Optional, Set

import pytest

from nextline import Nextline
from nextline.utils import agen_with_wait


this_dir = Path(__file__).resolve().parent


@pytest.fixture(autouse=True)
def monkey_patch_syspath(monkeypatch):
    monkeypatch.syspath_prepend(str(this_dir))
    yield


statement = this_dir.joinpath("script.py").read_text()


@pytest.mark.asyncio
async def test_run():
    nextline = Nextline(statement)
    assert nextline.state == "initialized"
    task_subscribe_state = asyncio.create_task(subscribe_state(nextline))
    task_control_execution = asyncio.create_task(control_execution(nextline))
    task_run = asyncio.create_task(run(nextline))
    await asyncio.gather(
        task_run, task_subscribe_state, task_control_execution
    )
    assert nextline.state == "closed"


async def subscribe_state(nextline: Nextline):
    return [s async for s in nextline.subscribe_state()]


async def run(nextline: Nextline):
    await asyncio.sleep(0.01)
    await nextline.run()
    nextline.exception()
    nextline.result()
    await nextline.close()


async def control_execution(nextline: Nextline):
    prev_ids: Set[int] = set()
    agen = agen_with_wait(nextline.subscribe_trace_ids())
    async for ids_ in agen:
        ids = set(ids_)
        new_ids, prev_ids = ids - prev_ids, ids

        tasks = {
            asyncio.create_task(control_trace(nextline, id_))
            for id_ in new_ids
        }
        _, pending = await agen.asend(tasks)

    await asyncio.gather(*pending)


async def control_trace(nextline: Nextline, trace_id):
    # print(f"control_trace({trace_id})")
    file_name = ""
    async for s in nextline.subscribe_prompting(trace_id):
        # print(s)
        if not file_name == s.file_name:
            file_name = s.file_name
            assert nextline.get_source(file_name)
        if s.prompting:
            command = "next"
            if s.trace_event == "line":
                line = nextline.get_source_line(
                    line_no=s.line_no,
                    file_name=s.file_name,
                )
                command = find_command(line) or command
            nextline.send_pdb_command(trace_id, command)


def find_command(line: str) -> Optional[str]:
    """The Pdb command indicated in a comment

    For example, returns "step" for the line "func()  # step"
    """
    import re

    if not (comment := extract_comment(line)):
        return None
    regex = re.compile(r"^# +(\w+) *$")
    match = regex.search(comment)
    if match:
        return match.group(1)
    return None


def extract_comment(line: str) -> Optional[str]:
    import io
    import tokenize

    comments = [
        val
        for type, val, *_ in tokenize.generate_tokens(
            io.StringIO(line).readline
        )
        if type == tokenize.COMMENT
    ]
    if comments:
        return comments[0]
    return None



from dataclasses import dataclass, field
from operator import itemgetter
from itertools import groupby
from keyword import iskeyword

from unittest.mock import Mock

from typing import Optional, Any, List, Tuple, Set, Dict
from types import FrameType


@dataclass
class TracedScope:
    module: List[str] = field(default_factory=list)
    func: List[str] = field(default_factory=list)


@dataclass
class TraceSummary:
    call: TracedScope = field(default_factory=TracedScope)
    line: TracedScope = field(default_factory=TracedScope)
    return_: TracedScope = field(default_factory=TracedScope)
    exception: TracedScope = field(default_factory=TracedScope)
    opcode: TracedScope = field(default_factory=TracedScope)


def summarize_trace_calls(
    mock_trace: Mock,
    modules: Optional[Set[str]] = None,
) -> TraceSummary:
    """Traced modules and functions for each event"""

    args_ = trace_call_args(mock_trace)  # type: ignore
    # [(frame, event, arg), ...]

    args = [
        {
            "event": event,
            "module": frame.f_globals.get("__name__"),
            "func": frame.f_code.co_name,
        }
        for frame, event, _ in args_
    ]
    # [{"event": event, "module": module, "func": func}, ...]

    if modules is not None:
        args = [d for d in args if d["module"] in modules]

    args = sorted(args, key=itemgetter("event"))
    # sorted for groupby()

    args_by_events = [
        (
            event + ("_" if iskeyword(event) else ""),  # "return_"
            list(args),  # expand to avoid exhaustion below
        )
        for event, args in groupby(args, itemgetter("event"))
    ]

    ret = TraceSummary(
        **{
            event: TracedScope(
                **{
                    scope: ordered_uniq_values(args, scope)
                    for scope in ("module", "func")
                }
            )
            for event, args in args_by_events
        }
    )
    return ret


def ordered_uniq_values(dicts: List[Dict], key: Any) -> List:
    """List of unique values of dicts for the key

    Note: The order is preserved while consecutive duplicates are removed.
    """
    return [v for v, _ in groupby([d[key] for d in dicts])]


def trace_call_args(trace: Mock) -> List[Tuple[FrameType, str, Any]]:
    return [call.args for call in trace.call_args_list]



from __future__ import annotations

import threading
import datetime

from typing import TYPE_CHECKING, Callable, Tuple, Any

from ..types import PromptInfo

if TYPE_CHECKING:
    from types import FrameType
    from pdb import Pdb
    from queue import Queue
    from ..utils import SubscribableDict


class PdbCommandInterface:
    """Relay pdb command prompts and commands

    An instance of this class is created for each execution of the pdb
    command loop, pdb._cmdloop().

    Parameters
    ----------
    pdb : Pdb
        The Pdb instance executing _cmdloop()
    queue_in : queue
        The queue connected to stdin in pdb
    queue_out : queue
        The queue connected to stdout in pdb

    """

    def __init__(
        self,
        pdb: Pdb,
        queue_in: Queue,
        queue_out: Queue,
        counter: Callable[[], int],
        trace_id: int,
        registry: SubscribableDict,
        trace_args: Tuple[FrameType, str, Any],
    ):
        self._pdb = pdb
        self._queue_in = queue_in
        self._queue_out = queue_out
        self._counter = counter
        self._trace_id = trace_id
        self._registry = registry
        self._ended = False
        self._nprompts = 0

        frame, event, _ = trace_args
        self._event = event
        self._file_name = pdb.canonic(frame.f_code.co_filename)
        self._line_no = frame.f_lineno
        self._prompt_no = -1

    def send_pdb_command(self, command: str) -> None:
        """send a command to pdb"""
        self._registry["prompt_info"] = PromptInfo(
            run_no=self._registry["run_no"],
            trace_no=self._trace_id,
            prompt_no=self._prompt_no,
            open=False,
            command=command,
            ended_at=datetime.datetime.now(),
        )
        self._command = command
        self._queue_in.put(command)

    def start(self) -> None:
        """start interfacing the pdb"""
        self._thread = threading.Thread(target=self._receive_pdb_stdout)
        self._thread.start()

    def end(self) -> None:
        """end interfacing the pdb"""
        self._ended = True
        self._queue_out.put(None)  # end the thread
        self._thread.join()

    def _receive_pdb_stdout(self):
        """receive stdout from pdb

        This method runs in its own thread during pdb._cmdloop()
        """
        while out := self._read_until_prompt(
            self._queue_out, self._pdb.prompt
        ):
            self._nprompts += 1
            self._prompt_no = self._counter()
            self._stdout = out
            self._registry["prompt_info"] = PromptInfo(
                run_no=self._registry["run_no"],
                trace_no=self._trace_id,
                prompt_no=self._prompt_no,
                open=True,
                event=self._event,
                file_name=self._file_name,
                line_no=self._line_no,
                stdout=out,
                started_at=datetime.datetime.now(),
            )

    def _read_until_prompt(self, queue, prompt):
        """read the queue up to the prompt"""
        out = ""
        while True:
            m = queue.get()
            if m is None:  # end
                return None
            out += m
            if prompt == m:
                break
        return out



import sys
import threading

from typing import Callable, Any, Type

from .types import TraceFunc

Func = Callable[[], Any]
DoneFunc = Callable[[Any, Type[Exception]], None]


def call_with_trace(
    func: Func,
    trace: TraceFunc,
    thread: bool = True,
) -> Any:
    """Set the trace function while running the function

    Notes
    -----
    The trace function will be used in all new threads created during the
    function execution regardless of whether the threads are created by the
    function. If the thread option is false, the trace function will not be use
    in any new threads.

    Parameters
    ----------
    func : callable
        A function to be called without any args. Use functools.partial to
        provide args. A return value and an exception will be given to `done`.
    trace: callable
        A trace function.
    thread: bool, default True
        If False, no new threads will be traced.

    Returns
    -------
    any
        The return value of func().

    Raises
    ------
    any
        An exception raised in func(). The exception is re-raised after the
        original trace function is put back.
    """

    ret = None
    exc = None

    trace_org = sys.gettrace()
    if thread:
        threading.settrace(trace)
    sys.settrace(trace)
    try:
        ret = func()
    except BaseException as e:
        exc = e
    finally:
        sys.settrace(trace_org)
        if thread:
            threading.settrace(trace_org)  # type: ignore
        if exc:
            raise exc
        return ret



import dataclasses
import datetime
from typing import Callable, Any, Optional, Union
from types import FrameType


ThreadNo = int
TaskNo = Union[int, None]


@dataclasses.dataclass(frozen=True)
class ThreadTaskId:
    thread_no: ThreadNo
    task_no: TaskNo


TraceFunc = Callable[
    [FrameType, str, Any], Optional[Callable[[FrameType, str, Any], Any]]
]
# Copied from (because not sure how to import)
# https://github.com/python/typeshed/blob/b88a6f19cdcf/stdlib/sys.pyi#L245


@dataclasses.dataclass(frozen=True)
class RunInfo:
    run_no: int
    state: str
    script: Optional[str] = None
    result: Optional[str] = None
    exception: Optional[str] = None
    started_at: Optional[datetime.datetime] = None
    ended_at: Optional[datetime.datetime] = None


@dataclasses.dataclass(frozen=True)
class TraceInfo:
    run_no: int
    state: str
    trace_no: int
    thread_no: int
    task_no: Optional[int] = None
    started_at: Optional[datetime.datetime] = None
    ended_at: Optional[datetime.datetime] = None


@dataclasses.dataclass(frozen=True)
class PromptInfo:
    run_no: int
    trace_no: int
    prompt_no: int
    open: bool
    event: Optional[str] = None
    file_name: Optional[str] = None
    line_no: Optional[int] = None
    stdout: Optional[str] = None
    command: Optional[str] = None
    started_at: Optional[datetime.datetime] = None
    ended_at: Optional[datetime.datetime] = None


@dataclasses.dataclass(frozen=True)
class StdoutInfo:
    run_no: int
    trace_no: int
    text: Optional[str] = None
    written_at: Optional[datetime.datetime] = None



import asyncio
import pytest

from nextline import Nextline

statement = """
import time

time.sleep(0.01)
"""


@pytest.mark.asyncio
async def test_run():

    nextline = Nextline(statement)
    assert nextline.state == "initialized"

    run = asyncio.create_task(nextline.run())

    await asyncio.sleep(0.01)

    async for state in nextline.subscribe_state():
        if state == "running":
            break

    async for trace_ids in nextline.subscribe_trace_ids():
        if trace_ids:
            trace_id = trace_ids[0]
            break

    n_prompting = 0
    async for state in nextline.subscribe_prompting(trace_id):
        if state.prompting:
            n_prompting += 1
            nextline.send_pdb_command(trace_id, "next")
    assert 3 == n_prompting

    await run
    assert nextline.state == "finished"
    await nextline.close()
    assert nextline.state == "closed"



from __future__ import annotations

import pytest
from typing import TYPE_CHECKING

from nextline.state import Initialized, Finished, Closed

if TYPE_CHECKING:
    from nextline.utils import SubscribableDict
    from nextline.state import State

from .base import BaseTestState

SOURCE_RAISE = """
raise Exception('foo', 'bar')
""".strip()

SOURCE_INVALID_SYNTAX = """
def
""".strip()


class TestFinished(BaseTestState):

    state_class = Finished

    @pytest.fixture()
    def state(self, finished):
        yield finished

    @pytest.mark.asyncio
    async def test_finish(self, state):
        # The same object should be returned no matter
        # how many times called.
        assert state is await state.finish()
        assert state is await state.finish()
        assert state is await state.finish()
        assert "obsolete" not in repr(state)

    @pytest.mark.asyncio
    async def test_reset(self, state):
        reset = state.reset()
        assert isinstance(reset, Initialized)
        await self.assert_obsolete(state)

    @pytest.mark.asyncio
    async def test_close(self, state):
        closed = state.close()
        assert isinstance(closed, Closed)
        await self.assert_obsolete(state)

    @pytest.mark.asyncio
    async def test_exception(self, state):
        assert state.exception() is None

    @pytest.mark.asyncio
    async def test_result(self, state):
        assert state.result() is None

    params = [
        pytest.param(SOURCE_RAISE, Exception, id="raise"),
        pytest.param(SOURCE_INVALID_SYNTAX, SyntaxError, id="invalid-syntax"),
    ]

    @pytest.mark.parametrize("source, exc", params)
    @pytest.mark.asyncio
    async def test_exception_raise(
        self, registry: SubscribableDict, source, exc
    ):
        registry["statement"] = source

        state: State = Initialized(registry=registry)
        state = state.run()

        state = await state.exited()
        state = await state.finish()
        assert isinstance(state, Finished)

        assert isinstance(state.exception(), exc)
        # assert ("foo", "bar") == state.exception().args
        with pytest.raises(exc):
            raise state.exception()

    @pytest.mark.parametrize("source, exc", params)
    @pytest.mark.asyncio
    async def test_result_raise(self, registry: SubscribableDict, source, exc):
        registry["statement"] = source

        state: State = Initialized(registry=registry)
        state = state.run()

        state = await state.exited()
        state = await state.finish()
        assert isinstance(state, Finished)

        with pytest.raises(exc):
            state.result()



from abc import ABC, abstractmethod
import itertools

import pytest

from nextline.state import Initialized, StateObsoleteError, StateMethodError
from nextline.utils import SubscribableDict, ThreadTaskIdComposer

SOURCE_ONE = """
import time
time.sleep(0.001)
""".strip()


class BaseTestState(ABC):
    """Test state classes of the state machine

    To be inherited by the test class for each state class.
    """

    @pytest.fixture()
    def statement(self):
        yield SOURCE_ONE

    @pytest.fixture()
    def registry(self, statement):
        y = SubscribableDict()
        y["statement"] = statement
        y["run_no_count"] = itertools.count().__next__
        y["trace_id_factory"] = ThreadTaskIdComposer()
        yield y
        y.close()

    @pytest.fixture()
    async def initialized(self, registry):
        y = Initialized(registry=registry)
        yield y
        if y.is_obsolete():
            return
        y.close()

    @pytest.fixture()
    async def running(self, initialized):
        y = initialized.run()
        yield y
        if y.is_obsolete():
            return
        exited = await y.exited()
        if exited.is_obsolete():
            return
        finished = await exited.finish()
        finished.close()

    @pytest.fixture()
    async def exited(self, running):
        y = await running.exited()
        yield y
        if y.is_obsolete():
            return
        finished = await y.finish()
        finished.close()

    @pytest.fixture()
    async def finished(self, exited):
        y = await exited.finish()
        yield y
        if y.is_obsolete():
            return
        y.close()

    @pytest.fixture()
    async def closed(self, finished):
        y = finished.close()
        yield y

    @abstractmethod
    def state(self, *_, **__):
        """Yield an instance of the class being tested

        To be overridden as a pytest fixture.
        """
        pass

    def test_state(self, state):
        assert isinstance(state, self.state_class)
        assert "obsolete" not in repr(state)

    async def assert_obsolete(self, state):
        assert "obsolete" in repr(state)

        with pytest.raises(StateObsoleteError):
            state.run()

        with pytest.raises(StateObsoleteError):
            await state.exited()

        with pytest.raises(StateObsoleteError):
            await state.finish()

        with pytest.raises(StateObsoleteError):
            state.reset()

        with pytest.raises(StateObsoleteError):
            state.close()

    def test_run(self, state):
        with pytest.raises(StateMethodError):
            state.run()

    @pytest.mark.asyncio
    async def test_exited(self, state):
        with pytest.raises(StateMethodError):
            await state.exited()

    @pytest.mark.asyncio
    async def test_finish(self, state):
        with pytest.raises(StateMethodError):
            await state.finish()

    @pytest.mark.asyncio
    async def test_reset(self, state):
        with pytest.raises(StateMethodError):
            state.reset()

    def test_send_pdb_command(self, state):
        thread_asynctask_id = (1, None)
        command = "next"
        with pytest.raises(StateMethodError):
            state.send_pdb_command(thread_asynctask_id, command)

    def test_exception(self, state):
        with pytest.raises(StateMethodError):
            state.exception()

    def test_result(self, state):
        with pytest.raises(StateMethodError):
            state.result()



"""The creation of a callable from code to be traced.

The name of this module becomes the module name of the callable.
"""

from functools import partial

from typing import Union, Callable
from types import CodeType


##__________________________________________________________________||
def compose(code: Union[CodeType, str]) -> Callable:
    """Create a function that executes the code

    Parameters
    ----------
    code : object
        A code to be executed when the returned function is called.
        This must be an object that can be executed by the Python
        built-in Function exec().

    Returns
    -------
    function
        A function with no arguments that executes the code

    """
    globals_ = {"__name__": __name__}
    # The globals, the second arg of exec(), resolves the issue*. An
    # empty dict would suffice for the issus.
    #
    # The __name__ is given for a different purpose. It becomes the
    # module name of the returned function. The module name is used in
    # Trace.
    #
    # [*] https://github.com/simonsobs/nextline/issues/7

    func = partial(exec, code, globals_)
    return func


##__________________________________________________________________||



import pytest
from unittest.mock import Mock

from typing import Set

from nextline.trace import Trace
from nextline.pdb.proxy import PdbInterfaceFactory, PdbInterface
from nextline.utils import SubscribableDict

from .funcs import TraceSummary

from . import module_a


def test_one(
    target: TraceSummary,
    probe: TraceSummary,
    ref: TraceSummary,
):
    assert ref.call.module
    assert ref.return_.module
    assert ref.call == target.call
    assert not target.return_.module

    assert ref == probe


def f():
    module_a.func_a()


@pytest.fixture()
def func():
    yield f


@pytest.fixture()
def thread():
    yield False


@pytest.fixture()
def target_trace_func(modules_to_trace: Set[str]):
    y = Trace(
        registry=Mock(spec=SubscribableDict),
        pdb_ci_map={},
        modules_to_trace=modules_to_trace,
    )
    yield y


@pytest.fixture()
def modules_to_trace():
    y = set()
    yield y


@pytest.fixture()
def mock_pdbi(probe_trace_func):
    y = Mock(spec=PdbInterface)
    y.trace = probe_trace_func
    yield y


@pytest.fixture(autouse=True)
def MockPdbInterfaceFactory(mock_pdbi, monkeypatch):
    mock_create_pdbi = Mock(return_value=mock_pdbi)
    y = Mock(spec=PdbInterfaceFactory, return_value=mock_create_pdbi)
    monkeypatch.setattr("nextline.trace.PdbInterfaceFactory", y)
    yield y



from __future__ import annotations

import sys
import io
import datetime
import linecache
from logging import getLogger

from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Optional,
    Tuple,
    TextIO,
)


from .state import Machine
from .utils import QueueDist, current_task_or_thread
from .types import StdoutInfo

if TYPE_CHECKING:
    from .pdb.proxy import PdbCIState
    from .types import RunInfo, TraceInfo, PromptInfo
    from .utils import SubscribableDict


class Nextline:
    """Nextline allows line-by-line execution of concurrent Python scripts

    Nextline supports concurrency with threading and asyncio. It uses multiple
    instances of Pdb, one for each thread and async task.

    Note
    ----
    The running asyncio event loop must exists when Nextline is instantiated.

    Parameters
    ----------
    statement : str
        A Python code as a string

    """

    def __init__(self, statement: str, run_no_start_from: int = 1):

        logger = getLogger(__name__)
        logger.debug(f"statement starts with {statement[:25]!r}")
        logger.debug(f"The next run number will be {run_no_start_from}")

        self.machine = Machine(statement, run_no_start_from)
        self.registry = self.machine.registry
        self._stdout = sys.stdout = IOSubscription(sys.stdout, self.registry)

    def __repr__(self):
        # e.g., "<Nextline 'running'>"
        return f"<{self.__class__.__name__} {self.state!r}>"

    async def run(self) -> None:
        """Execute the script and wait until it exits"""
        self.machine.run()
        await self.machine.finish()

    def exception(self) -> Optional[Exception]:
        """Uncaught exception from the last run"""
        return self.machine.exception()

    def result(self) -> Any:
        """Return value of the last run. always None"""
        return self.machine.result()

    def reset(
        self,
        statement: Optional[str] = None,
        run_no_start_from: Optional[int] = None,
    ) -> None:
        """Prepare for the next run"""
        self.machine.reset(
            statement=statement,
            run_no_start_from=run_no_start_from,
        )

    async def close(self) -> None:
        """End gracefully"""
        await self.machine.close()

    @property
    def statement(self) -> str:
        """The script"""
        return self.get("statement")

    @property
    def state(self) -> str:
        """The current condition of the script execution.

        The possible values are "initialized", "running", "exited", "finished",
        "closed"
        """
        return self.machine.state_name

    def subscribe_state(self) -> AsyncGenerator[str, None]:
        return self.subscribe("state_name")

    @property
    def run_no(self):
        """The current run number"""
        return self.get("run_no")

    def subscribe_run_no(self) -> AsyncGenerator[int, None]:
        return self.subscribe("run_no")

    def subscribe_trace_ids(self) -> AsyncGenerator[Tuple[int], None]:
        return self.subscribe("trace_nos")

    async def subscribe_prompting(
        self,
        trace_id: int,
    ) -> AsyncGenerator[PdbCIState, None]:
        agen = self.subscribe(trace_id)
        async for y in agen:
            if y is None:
                continue
            yield y

    def send_pdb_command(self, trace_id: int, command: str):
        self.machine.send_pdb_command(trace_id, command)

    def get_source(self, file_name=None):
        if not file_name or file_name == self.registry.get("script_file_name"):
            return self.get("statement").split("\n")
        return [e.rstrip() for e in linecache.getlines(file_name)]

    def get_source_line(self, line_no, file_name=None):
        """
        based on linecache.getline()
        https://github.com/python/cpython/blob/v3.9.5/Lib/linecache.py#L26
        """
        lines = self.get_source(file_name)
        if 1 <= line_no <= len(lines):
            return lines[line_no - 1]
        return ""

    def subscribe_run_info(self) -> AsyncGenerator[RunInfo, None]:
        return self.subscribe("run_info")

    def subscribe_trace_info(self) -> AsyncGenerator[TraceInfo, None]:
        return self.subscribe("trace_info")

    def subscribe_prompt_info(self) -> AsyncGenerator[PromptInfo, None]:
        return self.subscribe("prompt_info")

    def get(self, key) -> Any:
        return self.registry.get(key)

    def subscribe(self, key) -> AsyncGenerator[Any, None]:
        return self.registry.subscribe(key)

    def subscribe_stdout(self) -> AsyncGenerator[StdoutInfo, None]:
        return self._stdout.subscribe()


AGenDatetimeStr = AsyncGenerator[Tuple[datetime.datetime, str], None]


class IOSubscription(io.TextIOWrapper):
    def __init__(self, src: TextIO, registry: SubscribableDict):
        """Make output stream subscribable

        The src needs to be replaced with the instance of this class. For
        example, if the src is stdout,
            sys.stdout = IOSubscription(sys.stdout)

        NOTE: The code on the logic about the buffer copied from
        https://github.com/alphatwirl/atpbar/blob/894a7e0b4d81aa7b/atpbar/stream.py#L54
        """
        self.registry = registry
        self._queue: QueueDist[str] = QueueDist()
        self._src = src
        self._buffer = ""

        self._id_composer = registry["trace_id_factory"]  # type: ignore
        self._run_no_map = registry["run_no_map"]  # type: ignore
        self._trace_no_map = registry["trace_no_map"]  # type: ignore

    def write(self, s: str) -> int:

        ret = self._src.write(s)
        # TypeError if s isn't str as long as self._src is sys.stdout or
        # sys.stderr.

        if not self._id_composer.has_id():
            return ret

        self._buffer += s
        if s.endswith("\n"):
            self.flush()
        return ret

    def flush(self):
        if not self._buffer:
            return
        if not self._id_composer.has_id():
            return
        key = current_task_or_thread()
        if not (run_no := self._run_no_map.get(key)):
            return
        if not (trace_no := self._trace_no_map.get(key)):
            return
        now = datetime.datetime.now()
        info = StdoutInfo(
            run_no=run_no,
            trace_no=trace_no,
            text=self._buffer,
            written_at=now,
        )
        # print(info, file=sys.stderr)
        self._queue.put(info)
        self._buffer = ""

    async def subscribe(self):
        async for y in self._queue.subscribe(last=False):  # type: ignore
            yield y



import asyncio

import pytest

from nextline.utils import SubscribableDict, to_thread

from .aiterable import aiterable


@pytest.mark.asyncio
async def test_one():
    obj: SubscribableDict[str, str] = SubscribableDict()

    key = "key_one"
    items = ["item_one", "item_two", "item_three"]

    async def subscribe(last=True):
        return [y async for y in obj.subscribe(key, last=last)]

    async def send():
        async for i in aiterable(items):
            obj[key] = i
            assert i == obj[key]
        assert 1 == len(obj)
        assert [key] == list(obj)
        del obj[key]
        assert 0 == len(obj)
        assert [] == list(obj)

    results = await asyncio.gather(subscribe(), subscribe(False), send())
    # NOTE: the effect of the option `last` is not tested
    assert items == results[0]
    assert items == results[1]

    obj.close()


def test_key_error():
    obj = SubscribableDict()

    key = "no_such_key"
    with pytest.raises(KeyError):
        obj[key]

    assert 0 == len(obj)
    assert not obj

    with pytest.raises(KeyError):
        del obj[key]

    obj.close()


def test_get():
    obj = SubscribableDict()

    key = "no_such_key"
    assert obj.get(key) is None

    key = "no_such_key"
    assert 123 == obj.get(key, 123)

    assert 0 == len(obj)
    assert not obj

    obj.close()


nitems = [0, 1, 2, 50]
nsubscribers = [0, 1, 2, 70]


@pytest.mark.parametrize("nitems", nitems)
@pytest.mark.parametrize("nsubscribers", nsubscribers)
@pytest.mark.parametrize("thread", [True, False])
@pytest.mark.parametrize("close_register", [True, False])
@pytest.mark.asyncio
async def test_matrix(close_register, thread, nsubscribers, nitems):
    async def subscribe():
        return [y async for y in obj.subscribe(key)]

    def register():
        for item in items:
            obj[key] = item
            assert obj[key] == item
        if close_register:
            if key in obj:
                del obj[key]

    async def aregister():
        return register()

    obj = SubscribableDict()

    key = "item"
    items = [f"{key}-{i+1}" for i in range(nitems)]

    # subscribe
    tasks_subscribe = []
    for i in range(nsubscribers):
        task = asyncio.create_task(subscribe())
        tasks_subscribe.append(task)

    # register
    if thread:
        coro = to_thread(register)
    else:
        coro = aregister()
    task_register = asyncio.create_task(coro)

    await asyncio.gather(task_register)

    task_close = to_thread(obj.close)

    results = await asyncio.gather(*tasks_subscribe, task_close)
    expected = items
    actuals = results[:nsubscribers]
    for actual in actuals:
        assert actual == expected
